import enum
from dataclasses import dataclass
from datetime import date
from typing import List, NamedTuple


class DataSource(enum.StrEnum):
    big_query = "big_query"
    rpc = "rpc"
    chifra = "chifra"


class ChainType(enum.IntEnum):
    ethereum = 1
    polygon = 137
    optimism = 69
    arbitrum = 42161
    celo = 42220
    bsc = 56
    base = 8453
    avalanche = 43114


# closest: 'before' or 'after'
ChainTypeConfig = {
    ChainType.ethereum: {
        "allow": [DataSource.big_query, DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.etherscan.io/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0xc36442b4a4522e871399cd717abdd847ab11fe88",
        "aave_v3_pool_addr": "0x87870bca3f3fd6335c3f4ce8392d69350b4fa4e2",
        "squeeth_controller": "0x64187ae08781b09368e6253f9e94951243a493d5",
    },
    ChainType.polygon: {
        "allow": [DataSource.big_query, DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.polygonscan.com/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0xc36442b4a4522e871399cd717abdd847ab11fe88",
        "aave_v3_pool_addr": "0x794a61358d6845594f94dc1db02a252b5b4814ad",
    },
    ChainType.optimism: {
        "allow": [DataSource.big_query, DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api-optimistic.etherscan.io/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0xc36442b4a4522e871399cd717abdd847ab11fe88",
    },
    ChainType.arbitrum: {
        "allow": [DataSource.big_query, DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.arbiscan.io/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0xc36442b4a4522e871399cd717abdd847ab11fe88",
    },
    ChainType.celo: {
        "allow": [DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.celoscan.io/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0x3d79edaabc0eab6f08ed885c05fc0b014290d95a",
    },
    ChainType.bsc: {
        "allow": [DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.bscscan.com/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0x7b8a01b39d58278b5de7e48c8449c9f4f5170613",
    },
    ChainType.base: {
        "allow": [DataSource.rpc, DataSource.chifra],
        "query_height_api": "https://api.basescan.org/api?module=block&action=getblocknobytime&timestamp=%1&closest=%2",
        "uniswap_proxy_addr": "0x03a520b32c04bf3beef7beb72e919cf822ed34f1",
    },
    ChainType.avalanche: {
        "allow": [DataSource.big_query],
        "aave_v3_pool_addr": "0x794a61358D6845594F94dc1DB02A252b5b4814aD",
        "uniswap_proxy_addr": "0x655C406EBFa14EE2006250925e54ec43AD184f8B",
    },
}


class ToType(enum.StrEnum):
    minute = "minute"
    tick = "tick"
    raw = "raw"
    position = "position"
    user_lp = "user_lp"
    price = "price"


class ToFileType(enum.StrEnum):
    csv = "csv"
    feather = "feather"
    parquet = "parquet"


class DappType(enum.StrEnum):
    uniswap = "uniswap"
    aave = "aave"
    squeeth = "squeeth"


@dataclass
class BigQueryConfig:
    auth_file: str


@dataclass
class RpcConfig:
    end_point: str
    batch_size: int = 500
    auth_string: str | None = None
    keep_tmp_files: bool = False
    etherscan_api_key: str = None
    force_no_proxy: bool = False  # if set to true, will ignore proxy setting
    height_cache_path: str = None
    thread: int = 10


@dataclass
class ChifraConfig:
    etherscan_api_key: str = None  # query block number


@dataclass
class TokenConfig:
    name: str
    decimal: int


@dataclass
class UniswapConfig:
    pool_address: str
    ignore_position_id: bool = False  # if set to true, will not download proxy logs and leave a empty column
    token0: TokenConfig | None = None
    token1: TokenConfig | None = None
    is_token0_base: bool | None = None


class AaveKey(NamedTuple):
    day: date
    address: str


@dataclass
class AaveConfig:
    tokens: List[str]


@dataclass
class FromConfig:
    chain: ChainType
    data_source: DataSource
    dapp_type: DappType
    start: date
    end: date
    uniswap_config: UniswapConfig | None = None
    aave_config: AaveConfig | None = None
    big_query: BigQueryConfig | None = None
    chifra_config: ChifraConfig | None = None
    rpc: RpcConfig | None = None
    http_proxy: str | None = None


@dataclass
class ToConfig:
    type: ToType  # minute or tick
    save_path: str
    multi_process: bool = False
    skip_existed: bool = False
    keep_raw: bool = False
    to_file_type: ToFileType = ToFileType.csv


class KECCAK(enum.StrEnum):
    MINT = "0x7a53080ba414158be7ec69b987b5fb7d07dee101fe85488f0853ae16239d0bde"
    SWAP = "0xc42079f94a6350d7e6235f29174924f928cc2ac818eb64fed8004e115fbcca67"
    BURN = "0x0c396cd989a39f4459b5fa1aed6a9a8dcdbc45908acfd67e028cd568da98982c"
    COLLECT = "0x70935338e69775456a85ddef226c395fb668b63fa0115f5f20610b388e6ca9c0"
    UNI_PROXY_INCREASE = "0x3067048beee31b25b2f1681f88dac838c8bba36af25bfb2b7cf7473a5847e35f"
    UNI_PROXY_DECREASE = "0x26f6a048ee9138f2c0ce266f322cb99228e8d619ae2bff30c67f8dcf9d2377b4"
    UNI_PROXY_COLLECT = "0x40d0efd1a53d60ecbf40971b9daf7dc90178c3aadc7aab1765632738fa8b8f01"
    AAVE_SUPPLY = "0x2b627736bca15cd5381dcf80b0bf11fd197d01a037c52b927a881a10fb73ba61"
    AAVE_WITHDRAW = "0x3115d1449a7b732c986cba18244e897a450f61e1bb8d589cd2e69e6c8924f9f7"
    AAVE_BORROW = "0xb3d084820fb1a9decffb176436bd02558d15fac9b0ddfed8c465bc7359d7dce0"
    AAVE_REPAY = "0xa534c8dbe71f871f9f3530e97a74601fea17b426cae02e1c5aee42c96c784051"
    AAVE_LIQUIDATION = "0xe413a321e8681d831f4dbccbca790d2952b56f977908e45be37335533e005286"
    AAVE_UPDATED = "0x804c9b842b2748a22bb64b345453a3de7ca54a6ca45ce00d415894979e22897a"
    TRANSFER = "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
    SQUEETH_NORM_FACTOR_UPDATED = "0x339e53729b0447795ff69e70a74fed98fc7fef6fe94b7521099b32f0f8de4822"


uni_topic_mapping = {
    KECCAK.MINT.value: KECCAK.UNI_PROXY_INCREASE.value,  # mint
    KECCAK.SWAP.value: None,  # swap
    KECCAK.BURN.value: KECCAK.UNI_PROXY_DECREASE.value,  # burn
    KECCAK.COLLECT.value: KECCAK.UNI_PROXY_COLLECT.value,  # collect
}


@dataclass
class Config:
    from_config: FromConfig
    to_config: ToConfig
    id: str = ""


class MinuteData(object):
    def __init__(self):
        self.timestamp = None
        self.netAmount0 = 0
        self.netAmount1 = 0
        self.closeTick = None
        self.openTick = None
        self.lowestTick = None
        self.highestTick = None
        self.inAmount0 = 0
        self.inAmount1 = 0
        self.currentLiquidity = None

    def to_array(self):
        return [
            self.timestamp,
            self.netAmount0,
            self.netAmount1,
            self.closeTick,
            self.openTick,
            self.lowestTick,
            self.highestTick,
            self.inAmount0,
            self.inAmount1,
            self.currentLiquidity,
        ]

    def __repr__(self):
        return str(self.timestamp)

    def __str__(self):
        return str(self.timestamp)

    def fill_missing_field(self, prev_data) -> bool:
        """
        fill missing field with previous data
        :param prev_data:
        :return: data is available or not
        """
        if prev_data is None:
            prev_data = MinuteData()
        self.closeTick = self.closeTick if self.closeTick is not None else prev_data.closeTick
        self.openTick = self.openTick if self.openTick is not None else prev_data.closeTick
        self.lowestTick = self.lowestTick if self.lowestTick is not None else prev_data.closeTick
        self.highestTick = self.highestTick if self.highestTick is not None else prev_data.closeTick
        self.currentLiquidity = (
            self.currentLiquidity if self.currentLiquidity is not None else prev_data.currentLiquidity
        )

        return False if (self.closeTick is None or self.currentLiquidity is None) else True


MinuteDataNames = [
    "timestamp",
    "netAmount0",
    "netAmount1",
    "closeTick",
    "openTick",
    "lowestTick",
    "highestTick",
    "inAmount0",
    "inAmount1",
    "currentLiquidity",
]


class EthError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message


class NodeNames:
    uni_pool = "uni_pool"
    uni_proxy_transfer = "uni_proxy_transfer"
    uni_proxy_lp = "uni_proxy_LP"
    uni_minute = "uni_minute"
    uni_tick = "uni_tick"
    uni_tx = "uni_tx"
    uni_tick_without_pos = "uni_tick_without_pos"
    uni_positions = "uni_positions"
    uni_user_lp = "uni_user_lp"
    uni_relative_price = "uni_rel_price"

    aave_raw = "aave_raw"
    aave_minute = "aave_minute"
    aave_tick = "aave_tick"

    osqth_raw = "osqth_raw"
    osqth_minute = "osqth_minute"



import unittest

from demeter_fetch.processor_uniswap.uniswap_utils import x96_sqrt_to_decimal


# x96_sqrt_to_decimal


class UniUtilTest(unittest.TestCase):
    def test_x96_sqrt_to_decimal(self):
        # token0->usdc, token1->weth
        val = x96_sqrt_to_decimal(1438663542842353560857615249833810, 6, 18, False)
        self.assertEqual(val // 1, 3032.0)



import os
import time
import subprocess
from datetime import date, datetime
from typing import Dict

import pandas as pd

from demeter_fetch import ChainType
from demeter_fetch.common import ApiUtil, print_log
from demeter_fetch.sources.source_utils import get_height_from_date
from .source_utils import ContractConfig


def check_chifra() -> bool:
    """
    check either chifra exist and works.
    :return: true:exist, false: not exist
    """
    cmd = f"chifra --version"
    ex = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, shell=True, stderr=subprocess.PIPE
    )
    out, err = ex.communicate()
    status = ex.wait()
    if err is None:
        return False
    # get terimal encoding
    master, slave = os.openpty()
    encoding = os.device_encoding(master)
    std_str: str = err.decode(encoding)
    exist = std_str.startswith("chifra version")
    if not exist:
        raise RuntimeError("Chifra is not installed or not configured")


def get_chifra_cmd(start_height, end_height, contract, topic0, output_file_name) -> str:
    return f"chifra export --logs --fmt csv --first_block {start_height} --last_block {end_height} {contract} {topic0} > {output_file_name}"


def _join_topic(row):
    # generate less object to make it faster
    if pd.isna(row["topic3"]):
        if pd.isna(row["topic2"]):
            if pd.isna(row["topic1"]):
                return [row["topic0"]]
            else:
                return [row["topic0"], row["topic1"]]
        else:
            return [row["topic0"], row["topic1"], row["topic2"]]
    else:
        return [row["topic0"], row["topic1"], row["topic2"], row["topic3"]]


def query_log_by_chifra(
    chain: ChainType,
    day: date,
    contract: ContractConfig,
    to_path=".",
    keep_tmp_files: bool = False,
    http_proxy="",
    etherscan_api_key="",
) -> pd.DataFrame:
    check_chifra()
    print_log(f"Exporting logs in {day} from chifra")
    topic0 = ""
    if len(contract.topics) == 1:
        topic0 = contract.topics[0]
    tmp_file_name = os.path.join(
        to_path, f"{contract.address}-{topic0}-{day}.chifra.csv"
    )
    if not os.path.exists(tmp_file_name) or os.path.getsize(tmp_file_name) == 0:
        start_height, end_height = get_height_from_date(
            day, chain, http_proxy, etherscan_api_key
        )
        cmd = get_chifra_cmd(
            start_height, end_height, contract.address, topic0, tmp_file_name
        )
        subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()

    df = chifra_csv_to_raw_df(tmp_file_name, contract)
    if not keep_tmp_files:
        os.remove(tmp_file_name)
    return df


def chifra_csv_to_raw_df(path, contract: ContractConfig) -> pd.DataFrame:
    chifra_df = pd.read_csv(path)
    chifra_df = chifra_df[chifra_df["address"] == contract.address.lower()]
    chifra_df = chifra_df[chifra_df["topic0"].isin(contract.topics)]

    chifra_df = chifra_df.rename(
        columns={
            "blockNumber": "block_number",
            "date": "block_timestamp",
            "transactionHash": "transaction_hash",
            "transactionIndex": "transaction_index",
            "logIndex": "log_index",
            "data": "data",
        }
    )
    chifra_df["block_timestamp"] = chifra_df["block_timestamp"].apply(
        lambda x: x.replace(" UTC", "")
    )
    chifra_df["topics"] = chifra_df.apply(_join_topic, axis=1)
    raw_df = chifra_df[
        [
            "block_number",
            "block_timestamp",
            "transaction_hash",
            "transaction_index",
            "log_index",
            "topics",
            "data",
        ]
    ]

    return raw_df



from datetime import datetime
from typing import Callable, Dict, List

import pandas as pd

from demeter_fetch import Config
from demeter_fetch.common import DailyNode, DailyParam, get_tx_type, to_decimal, to_int, get_depend_name
from demeter_fetch.common import KECCAK, NodeNames
from .uniswap_utils import x96_sqrt_to_decimal

price_df_columns = ["block_timestamp", "token0", "token1", "total_liquidity", "sqrtPriceX96"]


class UniRelativePrice(DailyNode):
    """
    Will generate daily file.
    row is minute.
    column is token
    """

    name = NodeNames.uni_relative_price

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.rel-price"
            + self._get_file_ext()
        )

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]

    def set_config(self, config: Config):
        super().set_config(config)
        if config.from_config.uniswap_config.is_token0_base is None:
            raise RuntimeError("should set uniswap_config.is_token0_base")
        if config.from_config.uniswap_config.token0 is None:
            raise RuntimeError("should set uniswap_config.token0")
        if config.from_config.uniswap_config.token1 is None:
            raise RuntimeError("should set uniswap_config.token1")

    @property
    def _load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "token0": to_decimal,
            "token1": to_decimal,
            "total_liquidity": to_int,
            "sqrtPriceX96": to_int,
        }

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: datetime.date):
        df = data[get_depend_name(NodeNames.uni_tick_without_pos, self.id)]
        price_df = df[df["tx_type"] == "SWAP"].copy()
        if len(price_df.index) < 1:
            return pd.DataFrame(columns=price_df_columns)
        price_df.set_index("block_timestamp", inplace=True)
        price_df["price"] = price_df["sqrtPriceX96"].apply(
            lambda a: x96_sqrt_to_decimal(
                a,
                self.from_config.uniswap_config.token0.decimal,
                self.from_config.uniswap_config.token1.decimal,
                self.from_config.uniswap_config.is_token0_base,
            )
        )

        price_df = price_df[["price", "total_liquidity", "sqrtPriceX96"]]
        if self.from_config.uniswap_config.is_token0_base:
            price_df.rename(columns={"price": "token0"}, inplace=True)
            price_df["token1"] = 1
        else:
            price_df.rename(columns={"price": "token1"}, inplace=True)
            price_df["token0"] = 1

        new_index = pd.date_range(
            start=price_df.index[0].floor("D"),
            end=price_df.index[0].floor("D") + pd.Timedelta(days=1) - pd.Timedelta(minutes=1),
            freq="min",
        )
        price_df = price_df[~price_df.index.duplicated(keep="last")]
        price_df = price_df.resample("1min").last().ffill()  # resample to 1 min
        price_df = (
            price_df.reindex(new_index).ffill().bfill()
        )  # expend to whole day, and fill tail and head empty minutes
        price_df["block_timestamp"] = price_df.index
        price_df = price_df[price_df_columns]
        price_df.rename(
            columns={
                "token0": self.from_config.uniswap_config.token0.name,
                "token1": self.from_config.uniswap_config.token1.name,
            },
            inplace=True,
        )

        return price_df



from collections import namedtuple
from decimal import Decimal
from typing import Dict, Callable, List

import pandas as pd
from tqdm import tqdm

from .. import ChainType
from ..common import (
    ChainTypeConfig,
    NodeNames,
    Node,
    set_global_pbar,
    EmptyNamedTuple,
    to_decimal,
    get_depend_name,
)
from ..sources.rpc_utils import set_position_id


class UniUserLP(Node):
    name = NodeNames.uni_user_lp

    def _get_file_name(self, param: namedtuple) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-"
            f"{self.from_config.start.strftime('%Y-%m-%d')}-{self.from_config.end.strftime('%Y-%m-%d')}.user_lp"
            + self._get_file_ext()
        )

    @property
    def _load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "liquidity": to_decimal,
        }

    @property
    def _parse_date_column(self) -> List[str]:
        return ["start_time", "end_time"]

    def _process_one(self, data: Dict[str, List[str]], param: EmptyNamedTuple) -> pd.DataFrame():
        position_df = self.get_depend_by_name(NodeNames.uni_positions, self.id).read_file(
            data[get_depend_name(NodeNames.uni_positions, self.id)][0]
        )
        position_df = position_df[position_df["tx_type"].isin(["MINT", "BURN"])]
        position_df["liq_delta"] = position_df.apply(
            lambda r: -r["liquidity"] if r["tx_type"] == "BURN" else r["liquidity"], axis=1
        )
        grouped_position = position_df.groupby(["position_id"])
        user_lp_list = []
        for idx, position_tx in grouped_position:
            liquidity = Decimal(0)
            # before start time, if there are a mint, last liquidity might be negative
            # e.g. mint 10, before start time, then mint 20, and burn 30, We will only find mint 20 and burn 30,
            # liquidity at last will be -10
            # In this situation, we insert a moke mint transaction to add 10 liquidity.
            total_liq = position_tx["liq_delta"].sum()
            if total_liq < 0:
                user_lp_list.append(
                    {
                        "address": position_tx.iloc[0]["owner"],
                        "position_id": idx[0],
                        "tick_lower": position_tx.iloc[0]["tick_lower"],
                        "tick_upper": position_tx.iloc[0]["tick_upper"],
                        "start_time": None,
                        "end_time": position_tx.iloc[0]["block_timestamp"],
                        "liquidity": -total_liq,  # ,
                        "start_hash": "",
                        "end_hash": position_tx.iloc[0]["transaction_hash"],
                        "log_index": 0,
                        "block_number": position_tx.iloc[0]["block_number"] - 1,
                    }
                )
                liquidity = total_liq
            for i in range(len(position_tx.index)):
                if position_tx.iloc[i]["tx_type"] == "MINT":
                    liquidity += position_tx.iloc[i]["liquidity"]
                elif position_tx.iloc[i]["tx_type"] == "BURN":
                    liquidity -= position_tx.iloc[i]["liquidity"]
                    if liquidity <= 0:
                        continue
                row = {
                    "address": position_tx.iloc[i]["owner"],
                    "position_id": idx[0],
                    "tick_lower": position_tx.iloc[i]["tick_lower"],
                    "tick_upper": position_tx.iloc[i]["tick_upper"],
                    "start_time": position_tx.iloc[i]["block_timestamp"],
                    "end_time": None,
                    "liquidity": liquidity,  # ,
                    "start_hash": position_tx.iloc[i]["transaction_hash"],
                    "end_hash": "",
                    "log_index": position_tx.iloc[i]["pool_log_index"],
                    "block_number": position_tx.iloc[i]["block_number"],
                }
                if i < len(position_tx.index) - 1:
                    row["end_time"] = position_tx.iloc[i + 1]["block_timestamp"]
                    row["end_hash"] = position_tx.iloc[i + 1]["transaction_hash"]
                user_lp_list.append(row)
        df = pd.DataFrame(user_lp_list)
        df = df.sort_values(["address", "position_id", "block_number", "log_index"])
        df = df.drop(columns=["block_number", "log_index"])
        return df


class UniPositions(Node):
    name = NodeNames.uni_positions

    def _get_file_name(self, param: namedtuple) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-"
            f"{self.from_config.start.strftime('%Y-%m-%d')}-{self.from_config.end.strftime('%Y-%m-%d')}"
            f".positions" + self._get_file_ext()
        )

    @property
    def _load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "amount0": to_decimal,
            "amount1": to_decimal,
            "liquidity": to_decimal,
            "sqrtPriceX96": to_decimal,
        }

    def get_tx_user(self, chain: ChainType, txes: pd.DataFrame) -> str:
        """
        Get real owner of this tx by trace token flow
        """
        if len(txes.index) < 1:
            return ""
        tx: pd.Series = txes.iloc[0]

        if tx["to"] == ChainTypeConfig[chain]["uniswap_proxy_addr"]:
            return tx["from"]
        else:
            return tx["to"]

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]

    def _process_one(self, data: Dict[str, List[str]], param: EmptyNamedTuple) -> pd.DataFrame:
        tick_csv_paths = data[get_depend_name(NodeNames.uni_tick, self.id)]
        log_csv_paths = data[get_depend_name(NodeNames.uni_tx, self.id)]
        tick_csv_paths.sort()
        log_csv_paths.sort()
        pbar = tqdm(total=(self.from_config.end - self.from_config.start).days + 1, ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        total_df = pd.DataFrame()
        for i in range(len(tick_csv_paths)):
            daily_tick_df = self.get_depend_by_name(NodeNames.uni_tick).read_file(tick_csv_paths[i])
            daily_tx_df = self.get_depend_by_name(NodeNames.uni_tx).read_file(log_csv_paths[i])
            daily_tick_df = daily_tick_df[daily_tick_df["tx_type"].isin(["MINT", "BURN", "COLLECT"])]
            tx_hashes = daily_tick_df["transaction_hash"].drop_duplicates()
            owners = {
                hash: self.get_tx_user(self.from_config.chain, daily_tx_df[daily_tx_df["transaction_hash"] == hash])
                for hash in tx_hashes
            }
            daily_tick_df["owner"] = daily_tick_df["transaction_hash"].apply(lambda hash: owners[hash])
            total_df = pd.concat([total_df, daily_tick_df])
            pbar.update()

        # set position id for empty
        total_df["position_id"] = total_df.apply(set_position_id, axis=1)
        total_df = total_df[
            [
                "position_id",
                "tx_type",
                "owner",
                "tick_lower",
                "tick_upper",
                "liquidity",
                "block_number",
                "block_timestamp",
                "transaction_hash",
                "pool_tx_index",
                "pool_log_index",
                "proxy_log_index",
                "sender",
                "receipt",
                "amount0",
                "amount1",
                "sqrtPriceX96",
                "current_tick",
            ]
        ]
        total_df = total_df.sort_values(["position_id", "block_number", "pool_log_index"])
        return total_df



from decimal import Decimal

import numpy as np
import pandas as pd

import demeter_fetch.common._typing as _typing
from demeter_fetch.common import split_topic, get_tx_type


def signed_int(h):
    """
    Converts hex values to signed integers.
    """
    s = bytes.fromhex(h[2:])
    i = int.from_bytes(s, "big", signed=True)
    return i


def hex_to_address(topic_str):
    return "0x" + topic_str[26:]


def handle_proxy_event(topic_str):
    if topic_str is None or isinstance(topic_str, float) or len(topic_str) == 0:  # is none or nan
        return None
    topic_list = split_topic(topic_str)
    type_topic = topic_list[0]
    if len(topic_list) > 1 and (
        type_topic == _typing.KECCAK.UNI_PROXY_INCREASE.value
        or type_topic == _typing.KECCAK.UNI_PROXY_DECREASE.value
        or type_topic == _typing.KECCAK.UNI_PROXY_COLLECT.value
    ):
        return int(topic_list[1], 16)
    else:
        return None


def x96_sqrt_to_decimal(sqrt_priceX96, decimal0, decimal1, is_0_base):
    price = int(sqrt_priceX96) / 2**96
    tmp = (price**2) * (10 ** (decimal0 - decimal1))
    return tmp if is_0_base else 1 / tmp


def handle_event(tx_type, topics_str, data_hex):
    # proprocess topics string ->topic list
    # topics_str = topics.values[0]
    receipt = current_tick = tick_lower = tick_upper = None
    liquidity = Decimal(np.nan)
    delta_liquidity = Decimal(np.nan)
    current_liquidity = Decimal(np.nan)
    sqrtPriceX96 = Decimal(np.nan)

    topic_list = split_topic(topics_str)

    # data_hex = data.values[0]

    no_0x_data = data_hex[2:]
    chunk_size = 64
    chunks = len(no_0x_data)
    match tx_type:
        case _typing.KECCAK.SWAP:
            sender = hex_to_address(topic_list[1])
            receipt = hex_to_address(topic_list[2])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            amount0, amount1, sqrtPriceX96, current_liquidity, current_tick = [
                signed_int(onedata) for onedata in split_data
            ]

        case _typing.KECCAK.BURN:
            sender = hex_to_address(topic_list[1])
            tick_lower = signed_int(topic_list[2])
            tick_upper = signed_int(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            liquidity, amount0, amount1 = [signed_int(onedata) for onedata in split_data]
            delta_liquidity = -liquidity
        case _typing.KECCAK.MINT:
            # sender = topic_str_to_address(topic_list[1])
            owner = hex_to_address(topic_list[1])
            tick_lower = signed_int(topic_list[2])
            tick_upper = signed_int(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            sender = hex_to_address(split_data[0])
            liquidity, amount0, amount1 = [signed_int(onedata) for onedata in split_data[1:]]
            delta_liquidity = liquidity
        case _typing.KECCAK.COLLECT:
            tick_lower = signed_int(topic_list[2])
            tick_upper = signed_int(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            sender = hex_to_address(topic_list[1])
            receipt = hex_to_address(split_data[0])
            amount0, amount1 = [signed_int(onedata) for onedata in split_data[1:]]

        case _:
            raise ValueError("not support tx type")
    return (
        sender,
        receipt,
        Decimal(amount0),
        Decimal(amount1),
        Decimal(sqrtPriceX96),
        Decimal(current_liquidity),
        current_tick,
        tick_lower,
        tick_upper,
        Decimal(liquidity),
        Decimal(delta_liquidity),
    )


def add_proxy_log(df, index, proxy_row):
    df.loc[index, "proxy_data"] = proxy_row.data
    df.at[index, "proxy_topics"] = proxy_row.topics
    df.loc[index, "proxy_log_index"] = proxy_row.log_index


def match_proxy_log(pool_logs: pd.DataFrame, proxy_logs: pd.DataFrame):
    """

    :param pool_logs:
    :param proxy_logs:
    :return:
    """
    pool_logs["tx_type"] = pool_logs.apply(lambda x: get_tx_type(x.topics), axis=1)
    proxy_logs = proxy_logs.set_index(keys="transaction_hash")
    pool_logs["topics"] = pool_logs["topics"].apply(lambda x: split_topic(x))
    proxy_logs["topics"] = proxy_logs["topics"].apply(lambda x: split_topic(x))

    proxy_logs["topic_name"] = proxy_logs["topics"].apply(lambda x: x[0])
    pool_logs["topic_name"] = pool_logs["topics"].apply(lambda x: x[0])

    pool_logs["proxy_topics"] = [[]] * pool_logs.shape[0]

    for index, row in pool_logs.iterrows():
        if row.tx_type == _typing.KECCAK.SWAP:
            continue
        if row.transaction_hash not in proxy_logs.index:
            continue
        proxy_tx: pd.DataFrame = proxy_logs.loc[[row.transaction_hash]]
        proxy_tx_matched: pd.DataFrame = proxy_tx.loc[proxy_tx.topic_name == _typing.uni_topic_mapping[row.topic_name]]

        for pindex, possible_match in proxy_tx_matched.iterrows():
            if row.tx_type == _typing.KECCAK.MINT:
                if row.data[66:] == possible_match.data[2:]:
                    add_proxy_log(pool_logs, index, possible_match)
                    break
            elif row.tx_type == _typing.KECCAK.COLLECT or row.tx_type == _typing.KECCAK.BURN:
                if compare_burn_data(row.data, possible_match.data):
                    add_proxy_log(pool_logs, index, possible_match)
                    break
            else:
                raise ValueError("not supported tx type")
    # if no column is generated

    if "proxy_data" not in pool_logs.columns:
        pool_logs["proxy_data"] = None
        pool_logs["proxy_topics"] = [[]] * pool_logs.shape[0]
        pool_logs["proxy_log_index"] = None
    else:
        pool_logs["proxy_topics"] = pool_logs["proxy_topics"].fillna("[]")


def compare_int_with_error(a: int, b: int, error: int = None) -> bool:
    if error is None:
        if a > 10**10:
            error = 30
        elif a > 10**6:
            error = 15
        elif a > 10**2:
            error = 3
        else:
            error = 1
    return abs(a - b) <= error


def compare_burn_data(a: str, b: str) -> bool:
    """
    Compare burn topic data, to decide burn event in pool and proxy is the same.
    Amount in burn and collect might be different, so it has to compare_int_with_error
    example:
    1. https://polygonscan.com/tx/0x2d88b0cc9f8008135accc8667aa907931edf0be01d311fe437336be7cfe511fd#eventlog, there are two burn: log: 371,382
    2. https://polygonscan.com/tx/0xca50d94a36bc730a4ebb46b9e7535075d7da8a4efbbf9cc53638b058516dc907#eventlog, amount in collect topic is different

    Data struct:
    0x0000000000000000000000000000000000000000000000000014aca30ddf7569
      000000000000000000000000000000000000000000000000000041b051acc70d
      0000000000000000000000000000000000000000000000000000000000000000
    """
    if len(a) != 194 or len(b) != 194:
        return False
    if a[0:66] != b[0:66]:
        return False
    if not compare_int_with_error(int("0x" + a[66 : 66 + 64], 16), int("0x" + b[66 : 66 + 64], 16)):
        return False
    if not compare_int_with_error(int("0x" + a[66 + 64 : 66 + 2 * 64], 16), int("0x" + b[66 + 64 : 66 + 2 * 64], 16)):
        return False
    return True



from datetime import date

from demeter_fetch import (
    Config,
    ToConfig,
    FromConfig,
    ChainType,
    DataSource,
    DappType,
    ToType,
    ToFileType,
    RpcConfig,
)
from tests.test_squeeth_full import FullDownloadTest


class FullDownloadRPCTest(FullDownloadTest):

    def __init__(self, *args, **kwargs):
        super(FullDownloadRPCTest, self).__init__(*args, **kwargs)
        self.base_config = Config(
            FromConfig(
                chain=ChainType.ethereum,
                data_source=DataSource.rpc,
                dapp_type=DappType.squeeth,
                start=date(2024, 1, 5),
                end=date(2024, 1, 5),
                rpc=RpcConfig(
                    end_point=self.config["end_point"],
                    auth_string=self.config["auth"] if "auth" in self.config else None,
                    keep_tmp_files=False,
                    etherscan_api_key=self.config["etherscan_api_key"],
                    force_no_proxy=True,  # if set to true, will ignore proxy setting
                ),
                http_proxy=self.config["http_proxy"],
            ),
            ToConfig(
                type=ToType.raw,
                save_path=self.config["to_path"],
                skip_existed=True,
                to_file_type=ToFileType.csv,
                keep_raw=True,
                multi_process=False,
            ),
        )

    def test_raw(self):
        super().test_raw()

    def test_minute(self):
        super().test_minute()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-09 11:21
# @Author  : 32ethers
# @Description:
from collections import namedtuple
from datetime import timedelta, date

import os
from typing import List, Dict, Callable, Tuple

import pandas as pd
from tqdm import tqdm

from ._typing import Config, FromConfig, ToFileType
from .utils import TimeUtil, set_global_pbar, get_depend_name

EmptyNamedTuple = namedtuple("EmptyNamedTuple", [])


class Node:
    name = "ParentNode"

    def __init__(self):  # depends: List,
        self.id = ""
        self.depend_instance: List[Node] = []
        self.depends_dict: Dict[str, Node] = {}
        self.config: Config | None = None
        self.from_config: FromConfig | None = None
        self.to_path: str | None = None

    depend = []

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        if type(self) is not type(other):
            return False
        # !!! actually, instead of compare reference, i should compare property
        if self.config != other.config:
            return False
        return True

    def set_depend_instance(self, depends: List):
        self.depend_instance = depends
        self.depends_dict = {get_depend_name(d.name, d.id): d for d in self.depend_instance}

    def get_config_for_depend(self, depend_name: str) -> List[Config]:
        return [self.config]

    def set_config(self, config: Config):
        self.config = config
        self.from_config = config.from_config
        self.to_path = config.to_config.save_path

    def work(self):
        set_global_pbar(None)
        missing_params: List[EmptyNamedTuple] = []
        if self.config.to_config.skip_existed:
            step_file_names = self.get_file_paths
            for key, fn in step_file_names.items():
                if not os.path.exists(fn):
                    missing_params.append(key)
                    break
            if len(missing_params) < 1:
                return
        data = {}
        for depend in self.depend_instance:
            data[get_depend_name(depend.name, depend.id)] = list(depend.get_file_paths.values())
        pbar = tqdm(total=len(missing_params), ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        for param in missing_params:
            df = self._process_one(data, param)
            self.save_file(df, self.get_file_path(param))
            pbar.update()

    def _process_one(self, data: Dict[str, List[str]], param: namedtuple) -> pd.DataFrame:
        return pd.DataFrame()

    # region Function about files

    def _get_file_name(self, param: namedtuple) -> str:
        return f"{self.name}-{str(param)}.csv"

    def get_file_path(self, param: namedtuple) -> str:
        return os.path.join(self.to_path, self._get_file_name(param))

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        return {
            EmptyNamedTuple(): self.get_file_path(EmptyNamedTuple()),
        }

    @property
    def _load_csv_converter(self) -> Dict[str, Callable]:
        """
        if you want to read csv generated by this step with pandas, you have to use those csv converter
        :return:
        """
        return {}

    @property
    def _parse_date_column(self) -> List[str]:
        return []

    def _get_file_ext(self):
        match self.config.to_config.to_file_type:
            case ToFileType.csv:
                return ".csv"
            case ToFileType.feather:
                return ".feather"
            case ToFileType.parquet:
                return ".parquet"
            case _:
                raise RuntimeError(f"{self.config.to_config.to_file_type.name} not supported")

    def save_file(self, df: pd.DataFrame, path: str):
        match self.config.to_config.to_file_type:
            case ToFileType.csv:
                df.to_csv(path, index=False, lineterminator="\n", date_format="%Y-%m-%d %H:%M:%S")
            case ToFileType.feather:
                df.to_feather(path)
            case ToFileType.parquet:
                df.to_parquet(path, engine="pyarrow", index=False)
            case _:
                raise RuntimeError(f"{self.config.to_config.to_file_type.name} not supported")

    def read_file(self, path: str):
        match self.config.to_config.to_file_type:
            case ToFileType.csv:
                return pd.read_csv(path, converters=self._load_csv_converter, parse_dates=self._parse_date_column)
            case ToFileType.feather:
                return pd.read_feather(path)
            case ToFileType.parquet:
                return pd.read_parquet(path, engine="pyarrow")
            case _:
                raise RuntimeError(f"{self.config.to_config.to_file_type.name} not supported")

    def get_depend_by_name(self, depend_name: str, depend_id=""):
        return self.depends_dict[get_depend_name(depend_name, depend_id)]

    # endregion

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


DailyParam = namedtuple("DailyParam", ["day"])


class DailyNode(Node):
    """
    Node whose input and output and dependings are daily, and generate only one file per day.
    """

    def __init__(self):
        super().__init__()

    def work(self):
        set_global_pbar(None)
        # if daily, global loop will handle processbar, outfile existence, gather param
        day_idx = self.from_config.start
        pbar = tqdm(total=(self.from_config.end - self.from_config.start).days + 1, ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        while day_idx <= self.from_config.end:
            # day_str = day_idx.strftime("%Y-%m-%d")
            day_param = DailyParam(day_idx)
            step_file_name = self.get_file_path(day_param)
            if self.config.to_config.skip_existed and os.path.exists(step_file_name):
                day_idx += timedelta(days=1)
                pbar.update()
                continue
            param = {}
            for depend in self.depend_instance:
                depend_file_path = depend.get_file_path(day_param)
                param[get_depend_name(depend.name, depend.id)] = depend.read_file(depend_file_path)
            df = self._process_one_day(param, day_idx)
            self.save_file(df, self.get_file_path(day_param))
            day_idx += timedelta(days=1)
            pbar.update()

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date) -> pd.DataFrame:
        return pd.DataFrame()

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        return {
            DailyParam(day): self.get_file_path(DailyParam(day))
            for day in TimeUtil.get_date_array(self.from_config.start, self.from_config.end)
        }


AaveDailyParam = namedtuple("AaveDailyParam", ["day", "token"])


class AaveDailyNode(Node):
    """
    Node whose input and output and dependings are daily, and generate multiple files per day.
    """

    def __init__(
        self,
    ):
        super().__init__()

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        ret = {}
        for day in TimeUtil.get_date_array(self.from_config.start, self.from_config.end):
            for token in self.from_config.aave_config.tokens:
                param = AaveDailyParam(day, token)
                ret[param] = self.get_file_path(param)
        return ret

    def work(self):
        set_global_pbar(None)
        # if daily, global loop will handle processbar, outfile existence, gather param
        day_idx = self.from_config.start
        pbar = tqdm(total=(self.from_config.end - self.from_config.start).days + 1, ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        while day_idx <= self.from_config.end:
            # check file exist
            if self.config.to_config.skip_existed:
                all_exist = True
                for token in self.from_config.aave_config.tokens:
                    data_depends = AaveDailyParam(day_idx, token)
                    step_file_name = self.get_file_path(data_depends)
                    if not os.path.exists(step_file_name):
                        all_exist = False
                        break
                if all_exist:
                    day_idx += timedelta(days=1)
                    pbar.update()
                    continue

            data_depends = {}
            for depend in self.depend_instance:
                token_data = {}
                for token in self.from_config.aave_config.tokens:
                    path = depend.get_file_path(AaveDailyParam(day_idx, token))
                    token_data[token] = depend.read_file(path)
                data_depends[get_depend_name(depend.name, depend.id)] = token_data

            token_dfs = self._process_one_day(data_depends, day_idx, self.from_config.aave_config.tokens)
            for token, df in token_dfs.items():
                self.save_file(df, self.get_file_path(AaveDailyParam(day_idx, token)))
            day_idx += timedelta(days=1)
            pbar.update()

    def _process_one_day(
        self, data: Dict[str, Dict[str, pd.DataFrame]], day: date, tokens: List[str]
    ) -> Dict[str, pd.DataFrame]:
        return {}



import os
from datetime import date, timezone, datetime
from typing import List, Dict

import pandas as pd

import demeter_fetch.sources.rpc_utils as rpc_utils
from .source_utils import get_height_from_date
from .. import ChainType, ChainTypeConfig
from ..common import FromConfig, KECCAK, utils, split_topic, hex_to_length
from .source_utils import ContractConfig


def _update_df(df: pd.DataFrame) -> pd.DataFrame:
    if "block_timestamp" in df.columns:
        df["block_timestamp"] = df["block_timestamp"].apply(lambda x: x.replace("T", " "))

        columns = [
            "block_number",
            "block_timestamp",
            "transaction_hash",
            "transaction_index",
            "log_index",
            "topics",
            "data",
        ]
    else:
        columns = [
            "block_number",
            "transaction_hash",
            "transaction_index",
            "log_index",
            "topics",
            "data",
        ]
    df = df[columns]
    return df


# def query_tx_logs(
#     chain: ChainType,
#     end_point: str,
#     save_path: str,
#     auth_string: str | None = None,
#     http_proxy: str | None = None,
# ) -> pd.DataFrame:
#     client = rpc_utils.EthRpcClient(end_point, http_proxy, auth_string)
#     utils.print_log(f"Will download transaction logs of uniswap")
#
#     pass


def query_logs(
    chain: ChainType,
    end_point: str,
    save_path: str,
    start_height: int,
    end_height: int,
    contract: ContractConfig,
    batch_size: int = 500,
    auth_string: str | None = None,
    http_proxy: str | None = None,
    keep_tmp_files: bool = False,
    one_by_one: bool = False,
    skip_timestamp: bool = False,
    height_cache_path: str = None,
    thread: int = 10,
) -> pd.DataFrame:
    client = rpc_utils.EthRpcClient(end_point, http_proxy, auth_string)
    utils.print_log(f"Will download from height {start_height} to {end_height}")
    try:
        tmp_files_paths: List[str] = rpc_utils.query_event_by_height_concurrent(
            chain,
            client,
            contract,
            start_height,
            end_height,
            save_path=save_path,
            batch_size=batch_size,
            one_by_one=one_by_one,
            skip_timestamp=skip_timestamp,
            height_cache_path=height_cache_path,
            thread=thread,
        )
    except Exception as e:
        print(e)
        import traceback

        print(traceback.format_exc())
        exit(1)

    current_day_logs = []
    # Load temporary files based on height, then reorganize into raw files by day
    # Note: The logs in the tmp file have been sorted
    # utils.print_log("Generating daily files")
    for tmp_file in tmp_files_paths:
        logs: List[Dict] = rpc_utils.load_tmp_file(tmp_file)
        current_day_logs.extend(logs)
    df = pd.DataFrame(current_day_logs)
    if "block_dt" in df.columns:
        df = df.drop(columns=["block_dt"])
    if len(df.index) < 1:
        df = pd.DataFrame(
            columns=[
                "block_number",
                "block_timestamp",
                "transaction_hash",
                "transaction_index",
                "log_index",
                "topics",
                "data",
            ]
        )
    else:
        df = df.sort_values(["block_number", "log_index"], ascending=[True, True])
    # remove tmp files
    if not keep_tmp_files:
        for f in tmp_files_paths:
            if os.path.exists(f):
                os.remove(f)
    return df


def rpc_pool(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    start_height, end_height = get_height_from_date(day, config.chain, config.http_proxy, config.rpc.etherscan_api_key)
    daily_df = query_logs(
        chain=config.chain,
        end_point=config.rpc.end_point,
        save_path=save_path,
        start_height=start_height,
        end_height=end_height,
        contract=ContractConfig(
            config.uniswap_config.pool_address,
            [KECCAK.SWAP.value, KECCAK.BURN.value, KECCAK.COLLECT.value, KECCAK.MINT.value],
        ),
        batch_size=config.rpc.batch_size,
        auth_string=config.rpc.auth_string,
        http_proxy=config.http_proxy if not config.rpc.force_no_proxy else None,
        keep_tmp_files=config.rpc.keep_tmp_files,
        one_by_one=False,
        skip_timestamp=False,
        height_cache_path=config.rpc.height_cache_path,
        thread=config.rpc.thread,
    )
    daily_df = _update_df(daily_df)
    return daily_df


def rpc_proxy_lp(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    start_height, end_height = get_height_from_date(day, config.chain, config.http_proxy, config.rpc.etherscan_api_key)
    daily_df = query_logs(
        chain=config.chain,
        end_point=config.rpc.end_point,
        save_path=save_path,
        start_height=start_height,
        end_height=end_height,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["uniswap_proxy_addr"],
            [
                KECCAK.UNI_PROXY_DECREASE.value,
                KECCAK.UNI_PROXY_INCREASE.value,
                KECCAK.UNI_PROXY_COLLECT.value,
            ],
        ),
        batch_size=config.rpc.batch_size,
        auth_string=config.rpc.auth_string,
        http_proxy=config.http_proxy if not config.rpc.force_no_proxy else None,
        keep_tmp_files=config.rpc.keep_tmp_files,
        one_by_one=False,
        skip_timestamp=False,
        height_cache_path=config.rpc.height_cache_path,
        thread=config.rpc.thread,
    )
    daily_df = _update_df(daily_df)
    return daily_df


def rpc_proxy_transfer(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    start_height, end_height = get_height_from_date(day, config.chain, config.http_proxy, config.rpc.etherscan_api_key)
    daily_df = query_logs(
        chain=config.chain,
        end_point=config.rpc.end_point,
        save_path=save_path,
        start_height=start_height,
        end_height=end_height,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["uniswap_proxy_addr"],
            [KECCAK.TRANSFER.value],
        ),
        batch_size=config.rpc.batch_size,
        auth_string=config.rpc.auth_string,
        http_proxy=config.http_proxy if not config.rpc.force_no_proxy else None,
        keep_tmp_files=config.rpc.keep_tmp_files,
        one_by_one=True,
        skip_timestamp=True,
        height_cache_path=config.rpc.height_cache_path,
        thread=config.rpc.thread,
    )
    daily_df = _update_df(daily_df)
    return daily_df


def rpc_uni_tx(config: FromConfig, tx_hashes: pd.Series) -> pd.DataFrame:
    http_proxy = config.http_proxy if not config.rpc.force_no_proxy else None
    client = rpc_utils.EthRpcClient(config.rpc.end_point, http_proxy, config.rpc.auth_string)
    df = rpc_utils.query_tx(client, tx_hashes)
    # df = df.drop(columns=["from", "to"])
    return df


def rpc_aave(config: FromConfig, save_path: str, day: date, tokens):
    start_height, end_height = get_height_from_date(day, config.chain, config.http_proxy, config.rpc.etherscan_api_key)
    daily_df = query_logs(
        chain=config.chain,
        end_point=config.rpc.end_point,
        save_path=save_path,
        start_height=start_height,
        end_height=end_height,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["aave_v3_pool_addr"],
            [
                KECCAK.AAVE_REPAY.value,
                KECCAK.AAVE_BORROW.value,
                KECCAK.AAVE_SUPPLY.value,
                KECCAK.AAVE_WITHDRAW.value,
                KECCAK.AAVE_UPDATED.value,
                KECCAK.AAVE_LIQUIDATION.value,
            ],
        ),
        batch_size=config.rpc.batch_size,
        auth_string=config.rpc.auth_string,
        http_proxy=config.http_proxy if not config.rpc.force_no_proxy else None,
        keep_tmp_files=config.rpc.keep_tmp_files,
        one_by_one=False,
        skip_timestamp=False,
        height_cache_path=config.rpc.height_cache_path,
        thread=config.rpc.thread,
    )
    daily_df = _update_df(daily_df)
    daily_df["topics"] = daily_df["topics"].apply(lambda x: split_topic(x))
    daily_df["token"] = daily_df["topics"].apply(lambda r: hex_to_length(r[1], 40))
    daily_df = daily_df[daily_df["token"].isin(tokens)]
    return daily_df


def rpc_squeeth(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    if "squeeth_controller" not in ChainTypeConfig[config.chain]:
        raise RuntimeError(f"Squeeth does not exist in chain {config.chain.name}")
    start_height, end_height = get_height_from_date(day, config.chain, config.http_proxy, config.rpc.etherscan_api_key)
    daily_df = query_logs(
        chain=config.chain,
        end_point=config.rpc.end_point,
        save_path=save_path,
        start_height=start_height,
        end_height=end_height,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["squeeth_controller"],
            [KECCAK.SQUEETH_NORM_FACTOR_UPDATED.value],
        ),
        batch_size=config.rpc.batch_size,
        auth_string=config.rpc.auth_string,
        http_proxy=config.http_proxy if not config.rpc.force_no_proxy else None,
        keep_tmp_files=config.rpc.keep_tmp_files,
        one_by_one=True,
        skip_timestamp=True,
        height_cache_path=config.rpc.height_cache_path,
        thread=config.rpc.thread,
    )
    daily_df = _update_df(daily_df)
    daily_df["block_timestamp"] = daily_df["data"].apply(
        lambda x: datetime.fromtimestamp(int(x[64 * 3 + 2 :], 16), tz=timezone.utc)
    )

    return daily_df



import datetime
from dataclasses import dataclass
from typing import Dict, Callable, List

import pandas as pd

import demeter_fetch.processor_uniswap.uniswap_utils as uniswap_utils
from demeter_fetch.common import DailyNode, DailyParam, get_tx_type, get_depend_name
from demeter_fetch.common import KECCAK, NodeNames
from demeter_fetch.common import TextUtil, to_decimal

columns = [
    "timestamp",
    "netAmount0",
    "netAmount1",
    "closeTick",
    "openTick",
    "lowestTick",
    "highestTick",
    "inAmount0",
    "inAmount1",
    "currentLiquidity",
]


class ModuleUtils(object):
    @staticmethod
    def get_datetime(date_str: str) -> datetime:
        if isinstance(date_str, pd.Timestamp):
            return date_str.to_pydatetime()
        else:
            return datetime.datetime.strptime(
                TextUtil.cut_after(str(date_str), "+").replace("T", " "), "%Y-%m-%d %H:%M:%S"
            )


@dataclass
class MinuteData:
    timestamp = 0
    netAmount0 = 0
    netAmount1 = 0
    closeTick = 0
    openTick = 0
    lowestTick = 0
    highestTick = 0
    inAmount0 = 0
    inAmount1 = 0
    currentLiquidity = 0


class UniMinute(DailyNode):
    name = NodeNames.uni_minute

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.minute"
            + self._get_file_ext()
        )

    @property
    def _load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "inAmount0": to_decimal,
            "inAmount1": to_decimal,
            "currentLiquidity": to_decimal,
            "netAmount0": to_decimal,
            "netAmount1": to_decimal,
        }

    @property
    def _parse_date_column(self) -> List[str]:
        return ["timestamp"]

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: datetime.date) -> pd.DataFrame:
        df = data[get_depend_name(NodeNames.uni_pool, self.id)]
        if len(df.index) < 1:
            return pd.DataFrame(columns=columns)
        df["block_timestamp"] = pd.to_datetime(df["block_timestamp"])
        df = df.set_index(keys=["block_timestamp"])
        df["tx_type"] = df.apply(lambda x: get_tx_type(x.topics), axis=1)
        df = df[df["tx_type"] == KECCAK.SWAP]
        decoded_df = pd.DataFrame()
        decoded_df[
            [
                "sender",
                "receipt",
                "amount0",
                "amount1",
                "sqrtPriceX96",
                "currentLiquidity",
                "current_tick",
                "tick_lower",
                "tick_upper",
                "liquidity",
                "delta_liquidity",
            ]
        ] = df.apply(lambda r: uniswap_utils.handle_event(r.tx_type, r.topics, r.data), axis=1, result_type="expand")
        decoded_df["inAmount0"] = decoded_df["amount0"].apply(lambda x: x if x > 0 else 0)
        decoded_df["inAmount1"] = decoded_df["amount1"].apply(lambda x: x if x > 0 else 0)
        minute_df = decoded_df.resample("1min").agg(
            {
                "amount0": "sum",
                "amount1": "sum",
                "inAmount0": "sum",
                "inAmount1": "sum",
                "currentLiquidity": "last",
            }
        )
        minute_df = minute_df.rename(columns={"amount0": "netAmount0", "amount1": "netAmount1"})
        minute_df[["openTick", "highestTick", "lowestTick", "closeTick"]] = (
            decoded_df["current_tick"]
            .resample("1min")
            .agg(
                {
                    "openTick": "first",
                    "highestTick": "max",
                    "lowestTick": "min",
                    "closeTick": "last",
                }
            )
        )
        minute_df["timestamp"] = minute_df.index

        minute_df = minute_df[columns]
        minute_df[["closeTick", "currentLiquidity"]] = minute_df[["closeTick", "currentLiquidity"]].ffill()
        minute_df[["netAmount0", "netAmount1", "inAmount0", "inAmount1"]] = minute_df[
            ["netAmount0", "netAmount1", "inAmount0", "inAmount1"]
        ].fillna(value=0)
        minute_df["openTick"] = minute_df["openTick"].fillna(minute_df["closeTick"])
        minute_df["highestTick"] = minute_df["highestTick"].fillna(minute_df["closeTick"])
        minute_df["lowestTick"] = minute_df["lowestTick"].fillna(minute_df["closeTick"])

        return minute_df



import os
import unittest

import toml

from demeter_fetch import (
    ToType,
)
from demeter_fetch.core import download_by_config
from tests.utils import validate_files_by_md5


class AaveFullDownloadTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(AaveFullDownloadTest, self).__init__(*args, **kwargs)
        self.config = toml.load("config.toml")
        self.check_existed_result()
        self.base_config = None

    def check_existed_result(self):
        existed_files = os.listdir(self.config["to_path"])
        existed_files = list(filter(lambda f: f.endswith(".csv"), existed_files))
        if len(existed_files) > 0:
            is_del = input("Old result detects, delete them? y/n, default is n: ")
            if is_del == "y":
                [os.remove(os.path.join(self.config["uni_to_path"], f)) for f in existed_files]

    def test_raw(self):
        config = self.base_config
        generated_files = download_by_config(config)
        validate_files_by_md5(generated_files)

    def test_minute(self):
        config = self.base_config
        config.to_config.type = ToType.minute
        generated_files = download_by_config(config)
        validate_files_by_md5(generated_files)

    def test_tick(self):
        config = self.base_config
        config.to_config.type = ToType.tick
        generated_files = download_by_config(config)
        validate_files_by_md5(generated_files)



import time
from datetime import datetime

import demeter_fetch.common.utils as utils
from demeter_fetch import ChainType
from demeter_fetch.tools import bigquery_tools


def date_to_height(args):
    chain = ChainType[args.chain]
    start = datetime.strptime(args.start, "%Y-%m-%d").date()
    end = datetime.strptime(args.end, "%Y-%m-%d").date()
    start_height = utils.ApiUtil.query_blockno_from_time(
        chain,
        datetime.combine(start, datetime.min.time()),
        False,
        args.http_proxy,
        args.key,
    )
    sleep_time = 8
    if args.key is not None and args.key != "":
        sleep_time = 0.5
    utils.print_log(f"Querying end timestamp, wait for {sleep_time} seconds to prevent max rate limit")
    time.sleep(sleep_time)  # to prevent request limit
    end_height = utils.ApiUtil.query_blockno_from_time(
        chain,
        datetime.combine(end, datetime.max.time()),
        True,
        args.http_proxy,
        args.key,
    )
    print(f"Height range, start: {start_height}, end: {end_height}")


def block_timestamp_cache(args):
    chain = ChainType[args.chain]
    start_date = datetime.strptime(args.start, "%Y-%m-%d").date()
    end_date = datetime.strptime(args.end, "%Y-%m-%d").date()
    to_path = args.to
    proxy = args.http_proxy
    auth_path = args.key
    engine = args.engine
    bigquery_tools.get_block_and_timestamp_cache(auth_path, chain, to_path, start_date, end_date, proxy, engine)



from datetime import date

import pandas as pd

from .chifra_utils import query_log_by_chifra
from .source_utils import ContractConfig
from .. import ChainTypeConfig
from ..common import FromConfig
import demeter_fetch.common._typing as TYPE


def chifra_pool(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    df = query_log_by_chifra(
        chain=config.chain,
        day=day,
        contract=ContractConfig(
            config.uniswap_config.pool_address,
            [TYPE.KECCAK.SWAP.value, TYPE.KECCAK.BURN.value, TYPE.KECCAK.COLLECT.value, TYPE.KECCAK.MINT.value],
        ),
        to_path=save_path,
        keep_tmp_files=False,
        http_proxy=config.http_proxy,
        etherscan_api_key=config.chifra_config.etherscan_api_key,
    )
    return df


def chifra_proxy_lp(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    df = query_log_by_chifra(
        chain=config.chain,
        day=day,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["uniswap_proxy_addr"],
            [
                TYPE.KECCAK.UNI_PROXY_DECREASE.value,
                TYPE.KECCAK.UNI_PROXY_INCREASE.value,
                TYPE.KECCAK.UNI_PROXY_COLLECT.value,
            ],
        ),
        to_path=save_path,
        keep_tmp_files=False,
        http_proxy=config.http_proxy,
        etherscan_api_key=config.chifra_config.etherscan_api_key,
    )
    return df


def chifra_proxy_transfer(config: FromConfig, save_path: str, day: date) -> pd.DataFrame:
    df = query_log_by_chifra(
        chain=config.chain,
        day=day,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["uniswap_proxy_addr"],
            [TYPE.KECCAK.TRANSFER.value],
        ),
        to_path=save_path,
        keep_tmp_files=False,
        http_proxy=config.http_proxy,
        etherscan_api_key=config.chifra_config.etherscan_api_key,
    )
    return df


def chifra_aave(config: FromConfig, save_path: str, day: date, tokens) -> pd.DataFrame:
    df = query_log_by_chifra(
        chain=config.chain,
        day=day,
        contract=ContractConfig(
            ChainTypeConfig[config.chain]["aave_v3_pool_addr"],
            [
                TYPE.KECCAK.AAVE_REPAY.value,
                TYPE.KECCAK.AAVE_BORROW.value,
                TYPE.KECCAK.AAVE_SUPPLY.value,
                TYPE.KECCAK.AAVE_WITHDRAW.value,
                TYPE.KECCAK.AAVE_UPDATED.value,
                TYPE.KECCAK.AAVE_LIQUIDATION.value,
            ],
        ),
        to_path=save_path,
        keep_tmp_files=False,
        http_proxy=config.http_proxy,
        etherscan_api_key=config.chifra_config.etherscan_api_key,
    )
    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 12:04
# @Author  : 32ethers
# @Description:
import unittest
from typing import List

from demeter_fetch import (
    DappType,
    ToType,
    Config,
    ToConfig,
    FromConfig,
    ChainType,
    DataSource,
    UniswapConfig,
    TokenConfig,
)
from demeter_fetch.common import Node
from demeter_fetch.core import get_relative_nodes
from demeter_fetch.core.engine import get_root_node
from demeter_fetch.processor_squeeth import SqueethMinute
from demeter_fetch.processor_uniswap import UniTick, UniUserLP
from demeter_fetch.processor_uniswap.relative_price import UniRelativePrice


class TreeTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TreeTest, self).__init__(*args, **kwargs)

    def check_sequence(self, dapp: DappType, to_type: ToType, root_node, squence: List[str]):
        root: Node = get_root_node(dapp, to_type)
        root.set_config(
            Config(
                FromConfig(
                    chain=ChainType.ethereum,
                    data_source=DataSource.rpc,
                    dapp_type=DappType.uniswap,
                    start=None,
                    end=None,
                    uniswap_config=UniswapConfig("", False, TokenConfig("a", 6), TokenConfig("b", 6), True),
                ),
                ToConfig(
                    None,
                    "",
                    False,
                    False,
                    False,
                    None,
                ),
            ),
        )
        self.assertTrue(type(root) is root_node)
        result = get_relative_nodes(
            root,
        )
        result = [n.name for n in result]
        print(result)
        self.assertEqual(result, squence)

    def test_uni_tick(self):
        self.check_sequence(DappType.uniswap, ToType.tick, UniTick, ["uni_pool", "uni_proxy_LP", "uni_tick"])

    def test_price(self):
        self.check_sequence(
            DappType.uniswap, ToType.price, UniRelativePrice, ["uni_pool", "uni_tick_without_pos", "uni_rel_price"]
        )

    def test_user_lp(self):
        self.check_sequence(
            DappType.uniswap,
            ToType.user_lp,
            UniUserLP,
            ["uni_pool", "uni_proxy_LP", "uni_tick", "uni_tx", "uni_positions", "uni_user_lp"],
        )

    def test_osqth_minute(self):
        # requires two extra token price node
        self.check_sequence(
            DappType.squeeth,
            ToType.minute,
            SqueethMinute,
            [
                "osqth_raw",
                "uni_pool",
                "uni_tick_without_pos",
                "uni_rel_price",
                "uni_pool",
                "uni_tick_without_pos",
                "uni_rel_price",
                "osqth_minute",
            ],
        )



from typing import Dict, List

import pandas as pd

from demeter_fetch import KECCAK, NodeNames
from demeter_fetch.common import AaveDailyNode, get_tx_type, get_depend_name
from demeter_fetch.common.nodes import AaveDailyParam
from demeter_fetch.processor_aave.aave_utils import decode_event_ReserveDataUpdated
from datetime import datetime, date


class AaveMinute(AaveDailyNode):
    name = NodeNames.aave_minute

    def _get_file_name(self, param: AaveDailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-aave_v3-{param.token}-{param.day.strftime('%Y-%m-%d')}.minute"
            + self._get_file_ext()
        )

    def _process_one_day(self, data: Dict[str, Dict[str, pd.DataFrame]], day: date, tokens) -> Dict[str, pd.DataFrame]:
        ret: Dict[str, pd.DataFrame] = {}
        df = data[get_depend_name(NodeNames.aave_raw, self.id)]
        for token, token_df in df.items():
            token_df["tx_type"] = token_df.apply(lambda x: get_tx_type(x.topics), axis=1)
            token_df = token_df[token_df["tx_type"] == KECCAK.AAVE_UPDATED]
            ret[token] = preprocess_one(token_df)
        return ret

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]


def preprocess_one(raw_df: pd.DataFrame):
    result_df = pd.DataFrame()
    result_df[
        [
            "liquidity_rate",
            "stable_borrow_rate",
            "variable_borrow_rate",
            "liquidity_index",
            "variable_borrow_index",
        ]
    ] = raw_df.apply(decode_event_ReserveDataUpdated, axis=1, result_type="expand")
    result_df["block_timestamp"] = pd.to_datetime(raw_df["block_timestamp"])
    result_df = result_df.set_index("block_timestamp")
    if len(result_df.index) == 0:  # if empty
        return result_df
    # add start and end of the day, so after resample, there will always be 1440 row
    # please add tail first, because new line will always be added to tail.
    day_end = datetime.combine(result_df.index[0].date(), datetime.max.time(), result_df.index[0].tzinfo)
    result_df.loc[day_end] = result_df.tail(1).iloc[0]

    day_start = datetime.combine(result_df.index[0].date(), datetime.min.time(), result_df.index[0].tzinfo)
    if day_start not in result_df.index:
        result_df.loc[day_start] = result_df.head(1).iloc[0]
    result_df = result_df.resample("1min").last().ffill().shift(1).bfill()

    result_df.insert(loc=0, column="block_timestamp", value=result_df.index)

    return result_df



import argparse


def get_commend_args():
    argParser = argparse.ArgumentParser()
    argParser.add_argument("-c", "--config", help="Path of config file")
    parser_tool_sub = argParser.add_subparsers(help="demeter-fetch tools", dest="tools")

    parser_height_range = parser_tool_sub.add_parser(
        "date_to_height", help="Query block height from etherscan in certain date range"
    )
    parser_height_range.add_argument("-c", "--chain", help="chain name, [ethereum, polygon]")
    parser_height_range.add_argument("-s", "--start", help="start date, eg: 2023-1-1")
    parser_height_range.add_argument("-e", "--end", help="end date, eg: 2023-1-1")
    parser_height_range.add_argument(
        "-p", "--http_proxy", help="proxy, eg: https://localhost:7890, optional", default=""
    )
    parser_height_range.add_argument("-k", "--key", help="key in etherscan, optional", default="")

    block_timestamp = parser_tool_sub.add_parser(
        "block_timestamp",
        help="Generate a cache for block number and timestamp in leveldb, datasource is bigquery, and you need install plyvel with pip first",
    )
    block_timestamp.add_argument("-c", "--chain", help="chain name, [ethereum, polygon]")
    block_timestamp.add_argument("-s", "--start", help="start date, eg: 2023-1-1")
    block_timestamp.add_argument("-e", "--end", help="end date, eg: 2023-2-1")
    block_timestamp.add_argument("-p", "--http_proxy", help="proxy, eg: https://localhost:7890, optional", default="")
    block_timestamp.add_argument("-k", "--key", help="path of google bigquery api key")
    block_timestamp.add_argument("-t", "--to", help="save cache to this path")
    block_timestamp.add_argument("-n", "--engine", help="storage type, sqlite or levelDB", default="sqlite")

    args = argParser.parse_args()
    return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-10 10:44
# @Author  : 32ethers
# @Description:
from dataclasses import dataclass
from datetime import date
from typing import Dict, List

import pandas as pd

from .big_query import bigquery_aave, bigquery_pool, bigquery_proxy_lp, bigquery_proxy_transfer, bigquery_transaction
from .chifra import chifra_pool, chifra_proxy_lp, chifra_proxy_transfer, chifra_aave
from .rpc import rpc_pool, rpc_proxy_lp, rpc_proxy_transfer, rpc_uni_tx, rpc_aave, rpc_squeeth
from ..common import DataSource, NodeNames, DailyNode, DailyParam, AaveDailyNode, utils, get_depend_name
from ..common.nodes import AaveDailyParam


@dataclass
class EventLog:
    block_number = 0
    block_timestamp = 0
    transaction_hash = 0
    transaction_index = 0
    log_index = 0
    topics = 0
    data = 0


class UniSourcePool(DailyNode):
    name = NodeNames.uni_pool

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date):
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                df = bigquery_pool(self.from_config, day)
            case DataSource.rpc:
                df = rpc_pool(self.from_config, self.to_path, day)
            case DataSource.chifra:
                df = chifra_pool(self.from_config, self.to_path, day)
        return df

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.raw"
            + self._get_file_ext()
        )

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]


class UniSourceProxyLp(DailyNode):
    name = NodeNames.uni_proxy_lp

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date):
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                df = bigquery_proxy_lp(self.from_config, day)
            case DataSource.rpc:
                df = rpc_proxy_lp(self.from_config, self.to_path, day)
            case DataSource.chifra:
                df = chifra_proxy_lp(self.from_config, self.to_path, day)
        return df

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-uniswap-proxy-lp-{param.day.strftime('%Y-%m-%d')}.raw"
            + self._get_file_ext()
        )


class UniSourceProxyTransfer(DailyNode):
    name = NodeNames.uni_proxy_transfer

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date):
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                return bigquery_proxy_transfer(self.from_config, day)
            case DataSource.rpc:
                df = rpc_proxy_transfer(self.from_config, self.to_path, day)
            case DataSource.chifra:
                df = chifra_proxy_transfer(self.from_config, self.to_path, day)
        return df

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-uniswap-proxy-transfer-{param.day.strftime('%Y-%m-%d')}.raw"
            + self._get_file_ext()
        )

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]


class UniTransaction(DailyNode):
    name = NodeNames.uni_tx

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date):
        tick_df = data[get_depend_name(NodeNames.uni_tick, self.id)]
        tick_df = tick_df[tick_df["tx_type"].isin(["MINT", "BURN", "COLLECT"])]
        tx = tick_df["transaction_hash"].drop_duplicates()
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                df = bigquery_transaction(self.from_config, day, tx)
            case DataSource.rpc:
                df = rpc_uni_tx(self.from_config, tx)
            case DataSource.chifra:
                raise NotImplementedError()
        df.sort_values(by=["block_number", "transaction_index"], ascending=True, inplace=True)
        return df

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.tx.raw"
            + self._get_file_ext()
        )


class AaveSource(AaveDailyNode):
    name = NodeNames.aave_raw

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date, tokens: List[str]) -> Dict[str, pd.DataFrame]:
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                df = bigquery_aave(self.from_config, day, tokens)
            case DataSource.rpc:
                df = rpc_aave(self.from_config, self.to_path, day, tokens)
            case DataSource.chifra:
                df = chifra_aave(self.from_config, self.to_path, day, tokens)
        df["token"] = df["topics"].apply(lambda x: utils.hex_to_length(x[1], 40))
        tokens_df = {}
        for token_addr, token_df in df.groupby(["token"]):
            clean_df = token_df.drop(columns=["token"])
            tokens_df[token_addr[0]] = clean_df
        return tokens_df

    def _get_file_name(self, param: AaveDailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-aave_v3-{param.token}-{param.day.strftime('%Y-%m-%d')}.raw"
            + self._get_file_ext()
        )

    @property
    def _parse_date_column(self) -> List[str]:
        return ["block_timestamp"]


class SqueethSource(DailyNode):
    name = NodeNames.osqth_raw

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date):
        df: pd.DataFrame | None = None
        match self.from_config.data_source:
            case DataSource.big_query:
                # df = bigquery_pool(self.from_config, day)
                raise NotImplementedError()
            case DataSource.rpc:
                df = rpc_squeeth(self.from_config, self.to_path, day)
            case DataSource.chifra:
                # df = chifra_pool(self.from_config, self.to_path, day)
                raise NotImplementedError()

        return df

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-squeeth-controller-{param.day.strftime('%Y-%m-%d')}.raw"
            + self._get_file_ext()
        )



import os
from datetime import date, timedelta
from enum import Enum
from typing import List
from google.cloud import bigquery
from google.cloud.bigquery import Client

from demeter_fetch.common import print_log

global_client: Client | None = None


class BigQueryChain(Enum):
    ethereum = {
        "table_name": "bigquery-public-data.crypto_ethereum.logs",
        "tx_table_name": "bigquery-public-data.crypto_ethereum.transactions",
    }
    polygon = {
        "table_name": "public-data-finance.crypto_polygon.logs",
        "tx_table_name": "public-data-finance.crypto_polygon.transactions",
    }
    arbitrum = {
        "table_name": "bigquery-public-data.goog_blockchain_arbitrum_one_us.logs",
        "tx_table_name": "bigquery-public-data.goog_blockchain_arbitrum_one_us.transactions",
    }
    optimism = {
        "table_name": "bigquery-public-data.goog_blockchain_optimism_mainnet_us.logs",
        "tx_table_name": "bigquery-public-data.goog_blockchain_optimism_mainnet_us.transactions",
    }
    avalanche = {
        "table_name": "bigquery-public-data.goog_blockchain_avalanche_contract_chain_us.logs",
        "tx_table_name": "bigquery-public-data.goog_blockchain_avalanche_contract_chain_us.transactions",
    }


def _set_environment(auth_file: str, http_proxy: str = None):
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = auth_file
    if http_proxy:
        os.environ["http_proxy"] = http_proxy
        os.environ["https_proxy"] = http_proxy


def get_date_array(date_begin, date_end) -> List[date]:
    return [date_begin + timedelta(days=x) for x in range(0, 1 + (date_end - date_begin).days)]


def query_by_sql(query: str, auth_file: str, http_proxy: str | None = None):
    _set_environment(auth_file, http_proxy)
    global global_client
    if global_client is None:
        global_client = bigquery.Client()
    query_job = global_client.query(query)  # Make an API request.
    result = query_job.to_dataframe(create_bqstorage_client=False)
    return result


def close_client():
    global global_client
    if global_client is not None:
        global_client.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 14:22
# @Author  : 32ethers
# @Description:
from typing import List

from .. import DappType, ToType, Config
from ..common import Node
from ..processor_aave import AaveMinute, AaveTick
from ..processor_squeeth import SqueethMinute
from ..processor_uniswap import UniUserLP, UniPositions, UniTick, UniTickNoPos, UniMinute
from ..processor_uniswap.relative_price import UniRelativePrice
from ..sources import UniSourcePool, UniSourceProxyTransfer, UniSourceProxyLp, AaveSource, UniTransaction, SqueethSource


def _get_reversed_copy(list_to_reverse):
    ret_list = list_to_reverse.copy()
    ret_list.reverse()
    return ret_list


def get_relative_nodes(root: Node) -> List[Node]:
    depth_first_array = []
    stack = [root]
    while len(stack) > 0:
        current_node: Node = stack.pop()
        depth_first_array.append(current_node)
        current_node_depends = []
        for depend_class in current_node.__class__.depend:
            depend_configs: List = current_node.get_config_for_depend(depend_class.name)
            for depend_config in depend_configs:
                depend_instance = depend_class()
                depend_instance.set_config(depend_config)
                depend_instance.id = depend_config.id
                if depend_instance not in stack and depend_instance not in depth_first_array:
                    stack.append(depend_instance)
                current_node_depends.append(depend_instance)
        current_node.set_depend_instance(current_node_depends)
    depth_first_array.reverse()
    return depth_first_array


# region depends

# uniswap
UniSourcePool.depend = []
UniSourceProxyTransfer.depend = []
UniSourceProxyLp.depend = []
UniMinute.depend = [UniSourcePool]
UniTick.depend = [UniSourcePool, UniSourceProxyLp]
UniTickNoPos.depend = [UniSourcePool]
UniTransaction.depend = [UniTick]
UniPositions.depend = [UniTick, UniTransaction]
UniUserLP.depend = [UniPositions]
UniRelativePrice.depend = [UniTickNoPos]
# AAVE
AaveSource.depend = []
AaveMinute.depend = [AaveSource]
AaveTick.depend = [AaveSource]

# SQUEETH
SqueethSource.depend = []
SqueethMinute.depend = [SqueethSource, UniRelativePrice]


def get_root_node(dapp: DappType, to_type: ToType, ignore_pos_id: bool = False) -> Node:
    if dapp == DappType.uniswap:
        match to_type:
            case ToType.raw:
                return UniSourcePool()
            case ToType.tick:
                if ignore_pos_id:
                    return UniTickNoPos()
                else:
                    return UniTick()
            case ToType.position:
                return UniPositions()
            case ToType.minute:
                return UniMinute()
            case ToType.user_lp:
                return UniUserLP()
            case ToType.price:
                return UniRelativePrice()
            case _:
                raise NotImplemented(f"{dapp} {to_type} not supported")

    elif dapp == DappType.aave:
        match to_type:
            case ToType.raw:
                return AaveSource()
            case ToType.minute:
                return AaveMinute()
            case ToType.tick:
                return AaveTick()
            case _:
                raise NotImplemented(f"{dapp} {to_type} not supported")
    elif dapp == DappType.squeeth:
        match to_type:
            case ToType.raw:
                return SqueethSource()
            case ToType.minute:
                return SqueethMinute()
    else:
        raise NotImplemented(f"{dapp} not supported")



from datetime import date
from decimal import Decimal
from typing import List, Dict

import numpy as np
import pandas as pd

from demeter_fetch import NodeNames, Config, DappType, UniswapConfig, TokenConfig
from demeter_fetch.common import DailyNode, DailyParam, get_depend_name
import copy


class SqueethMinute(DailyNode):
    name = NodeNames.osqth_minute

    def _get_file_name(self, param: DailyParam) -> str:
        return (
            f"{self.from_config.chain.name}-squeeth-controller-{param.day.strftime('%Y-%m-%d')}.minute"
            + self._get_file_ext()
        )

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date) -> pd.DataFrame:
        squeeth_price_df = data[get_depend_name(NodeNames.uni_relative_price, "osqth-price")]
        squeeth_price_df = squeeth_price_df.set_index(["block_timestamp"])
        eth_price_df = data[get_depend_name(NodeNames.uni_relative_price, "eth-price")]
        eth_price_df = eth_price_df.set_index(["block_timestamp"])

        raw_df = data[get_depend_name(NodeNames.osqth_raw, self.id)]
        if len(raw_df.index) < 1:
            # in case this day is empty
            new_index = pd.date_range(
                start=pd.to_datetime(day).floor("D"),
                end=pd.to_datetime(day) + pd.Timedelta(days=1) - pd.Timedelta(minutes=1),
                freq="min",
            )
            df = pd.DataFrame(
                index=new_index,
                columns=["norm_factor", "WETH", "OSQTH"],
                data=np.nan,
            )
            df["WETH"] = eth_price_df["weth"]
            df["OSQTH"] = squeeth_price_df["osqth"]
            return df
        raw_df["block_timestamp"] = pd.to_datetime(raw_df["block_timestamp"].apply(lambda x: x[0:19]))
        raw_df = raw_df.set_index(["block_timestamp"])
        raw_df["oldNormFactor"] = raw_df["data"].apply(lambda x: Decimal(int(x[2 : 2 + 64], 16)) / Decimal(1e18))
        raw_df["newNormFactor"] = raw_df["data"].apply(
            lambda x: Decimal(int(x[2 + 64 : 2 + 64 * 2], 16)) / Decimal(1e18)
        )
        first_old_norm_factor = raw_df.iloc[0]["oldNormFactor"]
        new_index = pd.date_range(
            start=raw_df.index[0].floor("D"),
            end=raw_df.index[0].floor("D") + pd.Timedelta(days=1) - pd.Timedelta(minutes=1),
            freq="min",
        )
        raw_df.drop(
            columns=[
                "oldNormFactor",
                "data",
                "topics",
                "block_number",
                "transaction_hash",
                "transaction_index",
                "log_index",
            ],
            inplace=True,
        )
        raw_df = raw_df[~raw_df.index.duplicated(keep="last")]
        # resample to 1 min
        # 1. in every minute, use the last value of the minute
        # 2. fill empty minute with previous minute
        # 3. shift one minute, this will ensure value in 0 second is the right value
        #    (no transactions so it will be the same to the last minute, e.g. if tx at 8:00:15 make price to 1000,
        #    at 8:01:00 price should be 1000, regardless any transaction happens during 8:01)
        # 4. fill the first minute with the value of second minute
        raw_df = raw_df.resample("1min").last().ffill().shift(1).bfill()
        raw_df = raw_df.reindex(new_index).ffill()
        raw_df["newNormFactor"] = raw_df["newNormFactor"].fillna(value=first_old_norm_factor)
        raw_df["WETH"] = eth_price_df["weth"]
        raw_df["OSQTH"] = squeeth_price_df["osqth"]
        raw_df = raw_df.rename(columns={"newNormFactor": "norm_factor"})
        raw_df["block_timestamp"] = raw_df.index
        raw_df = raw_df[["block_timestamp", "norm_factor", "WETH", "OSQTH"]]
        return raw_df

    def get_config_for_depend(self, depend_name: str) -> List[Config]:
        match depend_name:
            case "uni_rel_price":
                osqth_eth_config: Config = copy.deepcopy(self.config)
                osqth_eth_config.id = "osqth-price"
                osqth_eth_config.from_config.dapp_type = DappType.uniswap
                osqth_eth_config.from_config.uniswap_config = UniswapConfig(
                    pool_address="0x82c427adfdf2d245ec51d8046b41c4ee87f0d29c",
                    token0=TokenConfig("weth", 18),
                    token1=TokenConfig("osqth", 18),
                    is_token0_base=True,
                )

                eth_usdc_config: Config = copy.deepcopy(self.config)
                eth_usdc_config.id = "eth-price"
                eth_usdc_config.from_config.dapp_type = DappType.uniswap
                eth_usdc_config.from_config.uniswap_config = UniswapConfig(
                    pool_address="0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8",
                    token0=TokenConfig("usdc", 6),
                    token1=TokenConfig("weth", 18),
                    is_token0_base=True,
                )
                return [osqth_eth_config, eth_usdc_config]
            case _:
                return [self.config]



import os
from datetime import date, datetime
from dateutil.relativedelta import relativedelta
from enum import Enum
from google.cloud import bigquery
from google.cloud.bigquery import Client
from typing import Tuple

from demeter_fetch import ChainType

global_client: Client | None = None


def _set_environment(auth_file: str, http_proxy: str = None):
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = auth_file
    if http_proxy:
        os.environ["http_proxy"] = http_proxy
        os.environ["https_proxy"] = http_proxy


class SupportedChain(Enum):
    arbitrum = "bigquery-public-data.goog_blockchain_arbitrum_one_us.blocks"
    optimism = "bigquery-public-data.goog_blockchain_optimism_mainnet_us.blocks"
    polygon = "bigquery-public-data.goog_blockchain_polygon_mainnet_us.blocks"
    ethereum = "bigquery-public-data.goog_blockchain_ethereum_mainnet_us.blocks"
    avalanche = "bigquery-public-data.goog_blockchain_avalanche_contract_chain_us.blocks"


class HeightCacheGenerator:
    def save(self, results):
        pass


class HeightCacheGeneratorLevelDB(HeightCacheGenerator):
    """
    height => block_timestamp cache
    """

    height_cache_file_name = "_height_timestamp_levelDB"

    def __init__(self, chain: ChainType, save_path: str):
        import plyvel

        self.height_cache_path = os.path.join(
            save_path, chain.name + HeightCacheGeneratorLevelDB.height_cache_file_name
        )
        self._block_dict = plyvel.DB(self.height_cache_path, create_if_missing=True)

    def save(self, results):
        for page in results.pages:
            with self._block_dict.write_batch() as wb:
                for row in page:
                    wb.put(row[0].to_bytes(4), int(row[1].timestamp() * 1000).to_bytes(6))

    def __del__(self):
        self._block_dict.close()


class HeightCacheGeneratorSqlite(HeightCacheGenerator):
    """
    height => block_timestamp cache
    """

    height_cache_file_name = "_height_timestamp.sqlite"

    def __init__(self, chain: ChainType, save_path: str):
        from sqlitedict import SqliteDict

        self.height_cache_path = os.path.join(save_path, chain.name + HeightCacheGeneratorSqlite.height_cache_file_name)
        self._block_dict = SqliteDict(self.height_cache_path, outer_stack=False)

    def save(self, results):
        in_mem_count = 0
        for page in results.pages:
            for row in page:
                self._block_dict[row[0]] = row[1]
                in_mem_count += 1
                if in_mem_count >= 1000:
                    self._block_dict.commit()
                    in_mem_count = 0
        self._block_dict.commit()

    def __del__(self):
        self._block_dict.close()


def _download(block_table: str, year: int, month: int, auth_path: str, proxy: str, days: Tuple[int, int] = None):
    if days is None:
        sql = (
            f"SELECT block_number,block_timestamp FROM `{block_table}` "
            f'WHERE TIMESTAMP_TRUNC(block_timestamp, MONTH) = TIMESTAMP("{year}-{month}-01")'
        )
    else:
        sql = (
            f"SELECT block_number,block_timestamp FROM `{block_table}` "
            f'WHERE DATE(block_timestamp) >= DATE("{year}-{month}-{days[0]}") '
            f'AND DATE(block_timestamp) <= DATE("{year}-{month}-{days[1]}");'
        )

    _set_environment(auth_path, proxy)
    global global_client
    if global_client is None:
        global_client = bigquery.Client()
    query_job = global_client.query(sql)  # Make an API request.
    result = query_job.result(page_size=1_000_000)
    return result


def download_between(
    start_date: date,
    end_date: date,
    chain: ChainType,
    google_auth_path: str,
    proxy: str,
    manager: HeightCacheGenerator,
):
    current_month = date(start_date.year, start_date.month, 1)

    while current_month <= end_date:
        print(datetime.now(), f"Querying {current_month.strftime('%Y-%m')}")
        big_query_iter = _download(
            SupportedChain[chain.name].value,
            current_month.year,
            current_month.month,
            google_auth_path,
            proxy,
        )
        print(datetime.now(), f"saving {current_month.strftime('%Y-%m')}")
        manager.save(big_query_iter)

        current_month = current_month + relativedelta(months=1)


def get_block_and_timestamp_cache(
    google_auth_path: str, chain: ChainType, to_path: str, start: date, end: date, proxy: str, engine: str
):
    print("===============preparing height cache from big query===============")

    if not os.path.exists(to_path):
        os.mkdir(to_path)
    if engine == "sqlite":
        mgr = HeightCacheGeneratorSqlite(chain, to_path)
    elif engine == "levelDB":
        mgr = HeightCacheGeneratorLevelDB(chain, to_path)
    else:
        raise RuntimeError(f"Unknown engine {engine}")
    download_between(start, end, chain, google_auth_path, proxy, mgr)
    print(datetime.now(), "saved to " + mgr.height_cache_path)



import os
import unittest
from datetime import datetime

import pandas as pd
import toml

import demeter_fetch.common._typing as typing
import demeter_fetch.sources.rpc_utils as rpc
from demeter_fetch.common import utils
from demeter_fetch.sources.source_utils import ContractConfig


class UniLpDataTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        self.config = toml.load("./config.toml")

        super(UniLpDataTest, self).__init__(*args, **kwargs)

    # ==========lines=========================
    def test_query_blockno_from_time(self):
        value = utils.ApiUtil.query_blockno_from_time(
            typing.ChainType.ethereum, datetime(2023, 5, 8), True, "127.0.0.1:7890"
        )
        self.assertTrue(value == 17212079, "height not right")

    def remove_tmp_file(self, paths: []):
        for p in paths:
            if os.path.exists(os.path.join(self.config["to_path"], p)):
                os.remove(os.path.join(self.config["to_path"], p))

    def test_query_event_by_height_concurrent(self):
        self.remove_tmp_file(
            [
                os.path.join(
                    self.config["to_path"],
                    "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42447801-42448800.tmp.pkl",
                )
            ]
        )
        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        height_cache = rpc.HeightCacheManager(typing.ChainType.polygon, self.config["to_path"])
        files = rpc.query_event_by_height_concurrent(
            chain=typing.ChainType.polygon,
            client=client,
            contract_config=ContractConfig(
                "0x45dda9cb7c25131df268515131f647d726f50608",
                [
                    typing.KECCAK.SWAP.value,
                    typing.KECCAK.BURN.value,
                    typing.KECCAK.COLLECT.value,
                    typing.KECCAK.MINT.value,
                ],
            ),
            start_height=42447801,
            end_height=42448800,  # diff = 999, will save in one batch
            height_cache=height_cache,
            save_path=self.config["to_path"],
            batch_size=500,
            skip_timestamp=False,
        )
        print(files)

    def test_query_event_by_height(self):
        self.remove_tmp_file(
            [
                os.path.join(
                    self.config["to_path"],
                    "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42447801-42448800.tmp.pkl",
                )
            ]
        )
        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        height_cache = rpc.HeightCacheManager(typing.ChainType.polygon, self.config["to_path"])
        files = rpc.query_event_by_height(
            chain=typing.ChainType.polygon,
            client=client,
            contract_config=ContractConfig(
                "0x45dda9cb7c25131df268515131f647d726f50608",
                [
                    typing.KECCAK.SWAP.value,
                    typing.KECCAK.BURN.value,
                    typing.KECCAK.COLLECT.value,
                    typing.KECCAK.MINT.value,
                ],
            ),
            start_height=42447801,
            end_height=42448800,  # diff = 999, will save in one batch
            height_cache=height_cache,
            save_path=self.config["to_path"],
            save_every_query=2,
            batch_size=500,
            skip_timestamp=True,
        )
        print(files)
        self.assertTrue(len(files) == 1)

    def test_query_event_by_height_save_rest(self):
        self.remove_tmp_file(
            [
                "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42447301-42448300.tmp.pkl",
                "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42448301-42448799.tmp.pkl",
            ]
        )

        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        files = self.query_3_save_2(client)
        print(files)
        self.assertTrue(len(files) == 2)

    def test_query_event_by_height_save_rest_again(self):
        self.remove_tmp_file(
            [
                "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42447301-42448300.tmp.pkl",
                "polygon-0x45dda9cb7c25131df268515131f647d726f50608-42448301-42448799.tmp.pkl",
            ]
        )

        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        files = self.query_3_save_2(client)
        print(files)
        self.assertTrue(len(files) == 2)
        # load from existing file again
        files = self.query_3_save_2(client)
        print(files)
        self.assertTrue(len(files) == 2)

    def query_3_save_2(self, client):
        return rpc.query_event_by_height(
            chain=typing.ChainType.polygon,
            client=client,
            contract_config=ContractConfig(
                "0x45dda9cb7c25131df268515131f647d726f50608",
                [
                    typing.KECCAK.SWAP.value,
                    typing.KECCAK.BURN.value,
                    typing.KECCAK.COLLECT.value,
                    typing.KECCAK.MINT.value,
                ],
            ),
            start_height=42447301,  # height difference cannot be divided by 2*500
            end_height=42448799,
            save_path=self.config["to_path"],
            save_every_query=2,
            batch_size=500,
            skip_timestamp=True,
        )

    def test_query_event_by_height_save_rest_remove_last(self):
        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        files = self.query_3_save_2(client)
        # just remove the last file.
        self.remove_tmp_file(["polygon-0x45dda9cb7c25131df268515131f647d726f50608-42448301-42448799.tmp.pkl"])
        files = self.query_3_save_2(client)
        print(files)
        self.assertTrue(len(files) == 2)

    def test_query_event_by_height_save_rest_remove_first(self):
        client = rpc.EthRpcClient(self.config["end_point"], "127.0.0.1:7890")
        files = self.query_3_save_2(client)
        # just remove the first file.
        self.remove_tmp_file(["polygon-0x45dda9cb7c25131df268515131f647d726f50608-42447301-42448300.tmp.pkl"])
        files = self.query_3_save_2(client)
        print(files)
        self.assertTrue(len(files) == 2)

    def test_query_tx_receipt(self):
        df = rpc.query_event_by_tx(
            rpc.EthRpcClient(self.config["end_point"]),
            pd.Series(
                [
                    "0xb4caa7d62ece248f8261d5e63ee76e69ed9fef0c9c72ea6cd33eae0ef9726512",
                    "0xea950a0f086d87175c4d3e7afa1299f80da9e32e2bb0460fef2aafc25e788afa",
                    "0x513de39afcc7be7cf8b3fb0391589d144c3de7c3f8f35ef0ea0317e98bc089e1",
                    "0x6e761fbdc3b0645d0f11181fd64cb1a73bc441c56716134826d67a0ad873f53d",
                    "0x20242b4d9db2d798945dcc3ddf55c3b98cc7a53c0898addd069d054f6b6867cb",
                    "0xef0419056caf269ee7b35b6ab2a8c966994206017cfd6d83a7010d888a307a75",
                    "0xeaa6ee98c343de36e96f257453a88fe2829c27ce4ef4ff0faffd9b66e5a025c1",
                    "0xeea6db2ffea4c06c665f5b902043521ea74be79fa490303fd996da6d9b1a2a71",
                    "0x06728af78607ee46346d74aa1ec967f3b7fea4b0a4c24cfb4db3bb31f2055cda",
                    "0xbfa9b92f0d91fc97c52f1986f478d282b56b3d9d898d613d0a6f5ec40122422d",
                    "0x0f92588945d6003724ce569a08854cc43334d1cf32e04c3b08f4dd1594abca81",
                    "0x3f77938b2bc9af2931652b02943935ad207548379f043890320ecbc4bfccd35d",
                    "0xe8b19fc85266a90050ddffffdd5b4d1ef57a5032cb3c1c75774d6a8f60e00682",
                    "0x8843edeacb1679d1ede456198c428634ee4089c54f934ac77eb4be5624843330",
                    "0xc6c9336989dc0a6f01cf3c976b6ffa59e8ce79213a05971a744962f43b6bc817",
                    "0xe0a25d2f322e71470545b07406fdbd867da4e3dfc51b65c55919f932f51316c7",
                    "0x92a49ecb2a738fb638afb6b9d8a88bcb99419a02224eac19c31be544600df4d4",
                    "0x9e1ab54c7a97cbbe149958e4122691eb382e7a59f0c184a258366436f4d220c8",
                    "0x0000000000000000000000000000000000000000000000000000000000000000",
                ]
            ),
            2,
        )
        self.assertTrue(len(df.index) == 182)

    def test_query_tx(self):
        df = rpc.query_tx(
            rpc.EthRpcClient(self.config["end_point"]),
            pd.Series(
                [
                    "0xb4caa7d62ece248f8261d5e63ee76e69ed9fef0c9c72ea6cd33eae0ef9726512",
                    "0xea950a0f086d87175c4d3e7afa1299f80da9e32e2bb0460fef2aafc25e788afa",
                    "0x513de39afcc7be7cf8b3fb0391589d144c3de7c3f8f35ef0ea0317e98bc089e1",
                    "0x6e761fbdc3b0645d0f11181fd64cb1a73bc441c56716134826d67a0ad873f53d",
                    "0x20242b4d9db2d798945dcc3ddf55c3b98cc7a53c0898addd069d054f6b6867cb",
                    "0xef0419056caf269ee7b35b6ab2a8c966994206017cfd6d83a7010d888a307a75",
                    "0xeaa6ee98c343de36e96f257453a88fe2829c27ce4ef4ff0faffd9b66e5a025c1",
                    "0xeea6db2ffea4c06c665f5b902043521ea74be79fa490303fd996da6d9b1a2a71",
                    "0x06728af78607ee46346d74aa1ec967f3b7fea4b0a4c24cfb4db3bb31f2055cda",
                    "0xbfa9b92f0d91fc97c52f1986f478d282b56b3d9d898d613d0a6f5ec40122422d",
                    "0x0f92588945d6003724ce569a08854cc43334d1cf32e04c3b08f4dd1594abca81",
                    "0x3f77938b2bc9af2931652b02943935ad207548379f043890320ecbc4bfccd35d",
                    "0xe8b19fc85266a90050ddffffdd5b4d1ef57a5032cb3c1c75774d6a8f60e00682",
                    "0x8843edeacb1679d1ede456198c428634ee4089c54f934ac77eb4be5624843330",
                    "0xc6c9336989dc0a6f01cf3c976b6ffa59e8ce79213a05971a744962f43b6bc817",
                    "0xe0a25d2f322e71470545b07406fdbd867da4e3dfc51b65c55919f932f51316c7",
                    "0x92a49ecb2a738fb638afb6b9d8a88bcb99419a02224eac19c31be544600df4d4",
                    "0x9e1ab54c7a97cbbe149958e4122691eb382e7a59f0c184a258366436f4d220c8",
                    "0x0000000000000000000000000000000000000000000000000000000000000000",
                ]
            ),
            2,
        )
        pass
        # self.assertTrue(len(df.index) == 182)



from setuptools import setup, find_packages

setup(
    name='demeter-fetch',
    version='1.1.10',
    packages=find_packages(exclude=["tests", "tests.*", "samples", "samples.*"]),
    url='',
    license='MIT',
    author='zelos research',
    author_email='zelos@antalpha.com',
    description='demeter fetch tool',
    python_requires='>=3.11',
    install_requires=[
        "numpy>=1.26.3",
        "pandas>=2.2.1",
        "protobuf>=4.25.2",
        "Requests>=2.31.0",
        "toml>=0.10.2",
        "tqdm>=4.66.2",
        "google-cloud-bigquery>=3.19.0",
        "db-dtypes>=1.2.0",
        "argparse>=1.4.0",
        "sqlitedict>=2.1.0",
    ],
    entry_points={
        'console_scripts': [
            'demeter-fetch = demeter_fetch.main:main',
        ],
    }
)
